package PodWiki::Runner;

$PodWiki::Runner::VERSION = "1.1";

use lib qw(. lib);

use strict;
no strict "refs";

use Exporter;

use Data::Dumper;
use Pod::Tree;
use CGI;
use CGI::Carp qw(fatalsToBrowser);
use CGI::Cookie;
use Digest::MD5  qw(md5_hex);
use MLDBM qw(DB_File Storable);
use DB_File;

use Rcs;
use IO::Socket;

use PodWiki::Config;

use vars qw(@ISA @EXPORT @EXPORT_OK @EXPORT_TAGS);
use vars qw(@head %seen %page_conf %users %groups %next %input %attributes %variables %index);
use vars qw($rcs $sub2go $query $me $current_user $page $output $break_mode $config $saving );

@ISA=qw(Exporter);
@EXPORT=qw();
@EXPORT_OK=qw();
@EXPORT_TAGS=();

my @docpages = qw(PodWikiHelp PodWikiEdit PodWikiDiff WikiWord
		 PodWikiUpload PodWikiRCS PodWikiAuth PerlPod PodWikiMarkup);

# the syntax map pod => html
my %syntax = (
	      head1       => "h1",
	      head2       => "h2",
	      head3       => "h3",
	      head4       => "h4",
	      text        => "p",
	      verbatim    => "pre",
	      b           => "b",
	      i           => "i",
	      c           => "code",
	      f           => "i",
	      list_number => "ol",
	      list_bullet => "ul",
	      list_text   => "dl",
	      item_number => "li",
	      item_bullet => "li",
	      item_text   => "dd",
	      );

sub new {
  my($this, %param) = @_;
  my $class = ref($this) || $this;
  my $self = \%param;
  bless($self, $class);
  return $self;
}

sub clean {
  # reset "global" VARS
  @head = %input = %seen = %page_conf = %users = %groups = %next = %attributes = %variables = %index = ();
  $config = $rcs = $sub2go = $query = $me = $current_user = $page = $output = $break_mode = $saving = "";
}

sub run {
  my($this) = @_;
  $this->clean();

  $me = $this->{me};

  $config = $this->{config};
  $config->fetch();

  # parse cgi input
  $query = new CGI;
  %input = map { $_ => $query->param($_) } $query->param;
  $page = $input{page} || "PodWikiIndex";
  $page =~ s/[`|\/\\\.\'\"]//g; # remove any unallowed characters

  # determine where to go
  $sub2go = $input{state} || "";

  if ($sub2go eq "imagetumb") {
    # binary output, do not run podwiki engine
    &imagetumb;
    exit;
  }

  # prepare rcs
  $rcs = &rcsinit;

  # before we go ahead, we check if authentication
  # is required
  &check_auth;

  # output is html
  print "Content-Type: text/html\n\n\n";

  eval {
    if (&data_empty) {
      # doing the initial setup
      $sub2go = "change_root_passwd";
      &setup;
    }
    &display;
  };
  if ($@) {
    die $@;
  }
}








sub display {
  # the core of the engine, fetch the page AutoLoad,
  # which itself includes all the stuff which drives
  # the piece.
  &tie_attributes($page);
  &render(&fetch_page("AutoLoad"));
  print $output;
}

sub edit {
  my ($content, $revision, $edit_title, $inherit, $attribute_source);
  $inherit = "";
  if ($input{preview}) {
    $content = $input{content};
    ##### $content =~ s/\r//g;
    print "<h3>Page Preview:</h3><div class=\"pdotted\">\n";
    &render($content);
    print "</div>";
  }
  else {
    if (-e "$config->{data}/$page") {
      $content = &fetch_page($page);
    }
  }

  if (-e "$config->{data}/$page") {
    $rcs->file($page);
    $revision = $rcs->head;
    $edit_title = "Edit Page";
    $attribute_source = $page;
  }
  else {
    $revision = "NEW";
    $edit_title = "Create Page";
    if (exists $input{inherit}) {
      # add note for editor
      $inherit = qq( properties inherited by
     <img src="images/ilink.png"> <a href="$me?page=$input{inherit}">$input{inherit}</a>);

      # where to fetch attributes from
      $attribute_source = $input{inherit};

      # use parent options, if any
      my %page_conf = &parse_options($attribute_source);
      if (scalar keys %page_conf) {
	# add this to the content
	$content  = "=options\n\n";
	$content .= join "\n", map { "$_ = $page_conf{$_}" } sort keys %page_conf;
	$content .= "\n\n=cut\n";
      }
    }
  }

  print qq(
<h3>$edit_title: <img src="images/ilink.png"> <i><a href="$me?page=$page">$page</a></i> (Revision: $revision)</h3>
$inherit
<p>Get some help about <img src="images/ilink.png"> <a href="$me?page=PodWikiMarkup">PodWikiMarkup</a>.
<form name="PodWiki" action="$me" method="post">
 <input type="hidden" name="page" value="$page">
 <input type="hidden" name="state" value="save">
 <textarea id="content" name="content" rows="25" cols="70">$content</textarea>
 <a href="javascript:ewiki_enlarge()" style="text-decoration:none">+</a>
 <script type="text/javascript">
   <!--
     function ewiki_enlarge() {
       var ta=document.getElementById("content");
       ta.style.width=((ta.cols*=1.0)*10).toString()+"px";
       ta.style.height=((ta.rows*=1.1)*20).toString()+"px";
     }
   //-->
 </script>
 <br>
 <table border="0">
  <tr><td align="right">Comment</td><td><input type="text" name="comment" class="dotted" size="70"></td></tr>
 </table>
);


  my $a_size = scalar keys %attributes;
  if ($a_size) {
    # attributes exists, so provide them
    print "<h4>Classify page:</h4>";
    print qq(<table border="0"><tr valign="top">);
    my $sel;
    foreach my $attribute (sort keys %attributes) {
      next if $attribute =~ /^\./;
      print qq(<td><b>$attribute</b><br>);
      print qq(<select name="podattributes-${attribute}" size="10"><option></option>);
      foreach my $attr_value (sort split /\s\s*/, $attributes{$attribute}) {
	if (exists $attributes{".classified"}->{$attribute}->{$attr_value}->{$attribute_source}) {
	  $sel = " selected";
	}
	else {
	  $sel = "";
	}
	print qq(<option${sel}>$attr_value</option>);
      }
      print qq(</select></td>);
    }
    print qq(</tr></table>);
  }

  print qq(
 <br><br>
 <input type="submit" name="save" value=" &nbsp; Save &nbsp; ">
 <input type="submit" name="preview" value=" &nbsp; Preview &nbsp; ">
 <a href="$me?page=$page">Cancel</a>
 </form>
 <hr noshade>
 <h3>Upload an image or another file:</h3>
 <form name=upload action="$me" method="post" enctype="multipart/form-data" >
   <input type=hidden name=state  value="upload">
   <input type=hidden name=page value="$page">
   <input type=file name=file size=40>
   <input type=submit name=submit value="upload this file">
 </form>
  );

  &fileindex;
}


sub edit_bar {
  print qq(<div align="right">);

  if ($current_user) {
    print qq(<font size="-2">logged in as
       <b><a href="$me?state=change_passwd">$current_user</a></b>
       </font>
        <a class="dotted" href="$me?state=logout&page=$page">logout</a>);
  }
  else {
    print qq(<a class="dotted" href="$me?state=login">login</a>);
  }

  if (! $sub2go || ($sub2go eq "save" && -e "$config->{data}/$page")) {
    $rcs->file($page);
    my $revision = $rcs->head;
    my $modified = scalar localtime($rcs->revdate);
    print qq(
       <font size="-2"><i>Last modified: $modified. Revision: $revision</i></font> 
      <a class="dotted" href="$me?page=$page&state=edit">EditThisPage</a> 
      <a class="dotted" href="$me?page=$page&state=info">InfoAboutPage</a>
    );
  }

  if (&is_member($current_user, "root")) {
    print qq(  <a class="dotted" href="$me?state=site_admin">SiteAdmin</a>);
  }

  print qq(</div>);
}


sub site_admin {
  print qq(
<h3>Welcome to the <I>PodWiki</I> Site-Admin Page</h3>

<table width="100%" border="0" cellspacing="10"><tr valign="top"><td >

<h4>Users</h4>
<form class="pdotted" name="newuser" action="$me" method="post">
<input type="hidden" name="state" value="new_user_admin">
New user: <input type="text" size="20" name="user">
<input type="submit" name="submit" value="create">
</form>
  );

  &tie_users;
  print qq(

 <table border="0" bgcolor="#000000" cellspacing="1" width=100%"><tr><td>
  <table border="0" cellspacing="1" cellpadding="5" width="100%">
   <tr bgcolor="#cfcfcf">
    <th>User</th><th>Action</th>
   </tr>
  );

  foreach my $user (sort keys %users) {
    my $home = "UserHome_${user}";
    my $link;
    if (-e "$config->{data}/$home") {
      $link = qq(<img src="images/ilink.png"> <a href="$me?page=$home">$user</a>);
    }
    else {
      $link = qq(<font class="nolink">$user</font><a href="$me?page=$home&state=edit">?</a>);
    }
    print qq(
    <tr bgcolor="#ffffff">
      <td>$link</td>
      <td>
        <a class="dotted" href="$me?page=$page&state=edit_user_admin&user=$user">edit</a>
        <a class="dotted" href="$me?page=$page&state=remove_user_admin&user=$user">remove</a>
      </td>
    </tr>
    );
  }

  print qq(</table></td></tr></table><p>&nbsp;</p>);



  print qq(
</td><td bgcolor="#ccccff"></td><td>

<h4>Groups</h4>
<form class="pdotted" name="newgroup" action="$me" method="post">
<input type="hidden" name="state" value="new_group_admin">
New group: <input type="text" size="20" name="group">
<input type="submit" name="submit" value="create">
</form>
  );

  &tie_groups;
  print qq(
 <table border="0" bgcolor="#000000" cellspacing="1" width=100%"><tr><td>
  <table border="0" cellspacing="1" cellpadding="5" width="100%">
   <tr bgcolor="#cfcfcf">
    <th>Group</th><th>Action</th>
   </tr>
  );

  foreach my $group (sort keys %groups) {
    print qq(
    <tr bgcolor="#ffffff">
      <td>$group</td>
      <td>
         <a class="dotted" href="$me?page=$page&state=edit_group_admin&group=$group">edit</a>
         <a class="dotted" href="$me?page=$page&state=remove_group_admin&group=$group">remove</a>
      </td>
    </tr>
    );
  }

  print qq(</table></td></tr></table></td></tr></table>);

  print qq(<h4>Attributes</h4>
<form class="pdotted" name="newattribute" action="$me" method="post">
<input type="hidden" name="state" value="new_attribute_admin">
New attribute: <input type="text" size="20" name="attribute">
<input type="submit" name="submit" value="create">
</form>
  );

  print qq(
 <table border="0" bgcolor="#000000" cellspacing="1" width=100%"><tr><td>
  <table border="0" cellspacing="1" cellpadding="5" width="100%">
   <tr bgcolor="#cfcfcf">
    <th>Attribute</th><th>Values (separated by whitespace)</th><th>Action</th>
   </tr>
  );

  foreach my $attribute ( sort keys %attributes) {
    next if $attribute =~ /^\./;
    print qq(
      <form name="change_${attribute}" action="$me" method="post">
      <input type="hidden" name="state" value="edit_attribute_admin">
      <input type="hidden" name="attribute" value="$attribute">
      <tr bgcolor="#ffffff">
        <td width="10%"><b>$attribute</b></td>
        <td width="70%">
         <input type="text" name="attr_values" size="60" value="$attributes{$attribute}">
        </td>
        <td>
          <input type="submit" name="submit" value="change">
          <a class="dotted" href="$me?page=$page&state=remove_attribute_admin&attribute=$attribute">remove</a>
        </td>
      </tr>
      </form>
      );
    }

  print qq(</table></td></tr></table>);

  &tie_variables;
    print qq(<h4>Variables</h4>
<form class="pdotted" name="newvariable" action="$me" method="post">
<input type="hidden" name="state" value="new_variable_admin">
New variable: <input type="text" size="20" name="variable">
<input type="submit" name="submit" value="create">
</form>
  );

    print qq(
 <table border="0" bgcolor="#000000" cellspacing="1" width=100%"><tr><td>
  <table border="0" cellspacing="1" cellpadding="5" width="100%">
   <tr bgcolor="#cfcfcf">
    <th>Variable</th><th>Value</th><th>Action</th>
   </tr>
  );

  foreach my $var (sort keys %variables) {
        print qq(
      <form name="change_${var}" action="$me" method="post">
      <input type="hidden" name="state" value="edit_variable_admin">
      <input type="hidden" name="variable" value="$var">
      <tr bgcolor="#ffffff">
        <td width="10%"><b>$var</b></td>
        <td width="70%">
         <input type="text" name="var_value" size="60" value="$variables{$var}">
        </td>
        <td>
          <input type="submit" name="submit" value="change">
          <a class="dotted" href="$me?page=$page&state=remove_variable_admin&variable=$var">remove</a>
        </td>
      </tr>
      </form>
      );
  }

  print qq(</table></td></tr></table>);

  print qq(<p>&nbsp;</p><p><a href="$me?state=docupdate_admin">Update PodWiki documentation</a></p>);
}

sub docupdate_admin {
  print qq(
     <h3>Update PodWiki documentation</h3>
     <form name="docupdate" method="post" action="$me">
        <input type="hidden" name="state" value="do_docupdate_admin">
	<select name="pages" size="10" multiple>
  );

  foreach my $docpage (@docpages) {
    print qq(<option selected>$docpage</option>);
  }

  print qq(
	</select>
      <br><input type="submit" name="submit" value="Update Pages from templates/ directory">
     </form>
  );
}

sub do_docupdate_admin {
  foreach my $file (@docpages) {
    system("cp $config->{templates}/$file $config->{data}/$file") and
      die "Could not install template file $file: $!\n";
    $rcs->file("$file");
    $rcs->ci('-l', "-m UpdatedByPodWikiSetup", "-wPodWiki(root)");
  }
  print qq(<p class="nolink">PodWiki documentation has been updated.</p>);
}

sub new_attribute_admin {
  if (! exists $attributes{$input{attribute}}) {
    # create it
    $attributes{$input{attribute}} = "";
  }
  else {
    print qq(<p class="nolink">Attribute <b>$input{attribute}</b> already exists!</p>);
  }
  &site_admin;
}

sub remove_attribute_admin {
  if (! exists $attributes{$input{attribute}}) {
    print qq(<p class="nolink">Attribute <b>$input{attribute}</b> does not exist!</p>);
  }
  else {
    delete $attributes{$input{attribute}};
    print qq(<p class="nolink">Attribute <b>$input{attribute}</b> has been removed.</p>);
  }
  &site_admin;
}

sub edit_attribute_admin {
  if (! exists $attributes{$input{attribute}}) {
    print qq(<p class="nolink">Attribute <b>$input{attribute}</b> does not exist!</p>);
  }
  else {
    my $values = join " ", sort split /\s\s*/, $input{attr_values};
    $attributes{$input{attribute}} = $values;
    print qq(<p class="nolink">Attribute <b>$input{attribute}</b> has been changed.</p>);
  }
  &site_admin;
}

sub new_variable_admin {
  if (! exists $variables{$input{variable}}) {
    # create it
    $variables{$input{variable}} = "";
  }
  else {
    print qq(<p class="nolink">Variable <b>$input{variable}</b> already exists!</p>);
  }
  &site_admin;
}

sub remove_variable_admin {
  if (! exists $variables{$input{variable}}) {
    print qq(<p class="nolink">Variable <b>$input{variable}</b> does not exist!</p>);
  }
  else {
    delete $variables{$input{variable}};
    print qq(<p class="nolink">Variable <b>$input{variable}</b> has been removed.</p>);
  }
  &site_admin;
}

sub edit_variable_admin {
  if (! exists $variables{$input{variable}}) {
    print qq(<p class="nolink">Variable <b>$input{variable}</b> does not exist!</p>);
  }
  else {
    $variables{$input{variable}} = $input{var_value};
    print qq(<p class="nolink">Variable <b>$input{variable}</b> has been changed.</p>);
  }
  &site_admin;
}

sub edit_user_admin {
  my($new) = @_;
  my $ro;
  if (! $new) {
    $ro = "readonly";
  }
  else {
    # new user
    &tie_users;
    if (exists $users{$input{user}}) {
      print qq(<p class="nolink">User <b>$input{user} already exists. Please try another username.</p>);
      $input{user} = "";
    }
  }
  my($user, $pw, $mail) = &fetch_user($input{user});
    print qq(
   <h3>Edit user $input{user}</h3>
   <form name="edituser" action="$me" method="post">
   <input type="hidden" name="state" value="do_save_user_admin">
    <table border="0">
      <tr>
        <th align="right">User</th>
	<td><input type="text" size="20" name="user" value="$input{user}" $ro></td>
      </tr>
      <tr>
        <th align="right">Email</th>
	<td><input type="text" size="20" name="mail" value="$mail"></td>
      </tr>
      <tr>
        <th align="right">New password</th>
	<td><input type="password" size="20" name="passwd"></td>
      </tr>
      <tr>
        <td>&nbsp;</td>
        <td><input type="submit" name="submit" value="Save"></td>
      </tr>
    </table>
   </form>
  );
}

sub new_user_admin {
  &edit_user_admin("new");
}

sub do_save_user_admin {
  if ($input{user} =~ /[`|\/\\\.\'\"]/) {
    &edit_user_admin("new");
    print qq(<p class="nolink">Username contained forbidden characters. Please try another username.</p>);
    return;
  }
  else {
    # check its existence
    &tie_users;
    if (exists $users{$input{user}}) {
      &edit_user_admin("new");
      return;
    }
  }

  if (($input{passwd} || $input{mail}) && $input{user}) {
    &save_user($input{user}, $input{passwd}, $input{mail});
    &site_admin;
    print qq(<p class="nolink">User <b>$input{user}</b> saved</p>);
    return;
  }
  else {
    &site_admin;
    print qq(<p class="nolink">Incomplete input</p>);
    return;
  }
}

sub remove_user_admin {
  if ($input{user}) {
    &del_user($input{user});
    &del_member($input{user}, "*");
    &site_admin;
    print qq(<p class="nolink">User <b>$input{user}</b> removed</p>);
  }
}


sub edit_group_admin {
  my($new) = @_;
  my $ro;
  if (! $new) {
    $ro = "readonly";
  }

  print qq(
   <h3>Edit group $input{group}</h3>
   <form name="editgroup" action="$me" method="post">
   <input type="hidden" name="state" value="do_save_group_admin">
   <input type="hidden" name="group" value="$input{group}" $ro>
    <table border="0">
      <tr>
        <th align="right">Group</th>
	<td><input type="text" name="group" value="$input{group}" $ro></td>
      </tr>
      <tr valign="top">
        <th align="right">Members</th>
	<td>
	  <select name="members" size="10" multiple>
    );

    &tie_users;
    foreach my $user (sort keys %users) {
      next if(! $user);
      if (! $new) {
	if (&is_member($user, $input{group})) {
	  print qq(      <option selected>$user</option>);
	}
	else {
	  print qq(      <option>$user</option>);
	}
      }
      else {
	print qq(      <option>$user</option>);
      }
    }

    print qq(	  </select>
      <br><font size="-1">Select multiple users by pressing CTRL while clicking on user names</font>
        </td>
      </tr>
      <tr>
        <td>&nbsp;</td>
        <td><input type="submit" name="submit" value="Save"></td>
      </tr>
    </table>
   </form>
  );
}

sub remove_group_admin {
  if ($input{group}) {
    &del_group($input{group});
    &site_admin;
    print qq(<p class="nolink">Group <b>$input{group}</b> removed</p>);
  }
}

sub new_group_admin {
  &edit_group_admin("new");
}

sub do_save_group_admin {
  my @members = $query->param("members");  # shall be an array
  if ($input{group}) {
    &set_members($input{group}, @members); # creates the group
    &site_admin;
    print qq(<p class="nolink">Group <b>$input{group}</b> saved</p>);
  }
  else {
    &site_admin;
    print qq(<p class="nolink">Incomplete input</p>);
    return;
  }
}


sub toc {
  # generate a table of content
  my $space;
  print "<p>\n";
  foreach my $target (@head) {
    $space = "&nbsp;&nbsp;&nbsp;&nbsp;" x $target->{level};
    print qq( $space<a href="#$target->{target}">$target->{title}</a><br>);
  }
  print "</p><p>&nbsp;</p>\n";
}

sub toc_by_attribute {
  # generate a table of content by selected attribute
  # FIXME: support for toc by multiple attributes!!!
  my ($attribute, $value);
  foreach my $attr (keys %attributes) {
    next if $attribute =~ /^\./;
    if (exists $input{$attr}) {
      # this one has been set by CGI
      $attribute = $attr;
      $value     = $input{$attr};
    }
  }
  print "<p>\n";
  if (exists $attributes{".classified"}->{$attribute}->{$value}) {
    foreach my $page (sort keys %{$attributes{".classified"}->{$attribute}->{$value}}) {
      print qq(&nbsp;&nbsp;&nbsp;&nbsp;<img src=\"images/ilink.png\"> <a href="$me?page=$page&$attribute=$value">$page</a><br>);
    }
  }
  print "</p>\n";
}

sub clean_head {
  my $head = shift;
  $head =~ s/[><\.\_\-\(\)\/\\\#\:\;\"\'\s]//g;
  return $head;
}

sub fetch_toc {
  my $content = shift;
  local $_;
  @head = ();
  foreach (split /\n/, $content) {
    chomp;
    if (/^=head(\d) (.*$)$/) {
      my $level  = $1;
      my $target = $2;
      push @head, {
		   level  => $level,
		   target => &clean_head($target),
		   title  => $target
		   };
    }
  }
}

sub fetch_page {
  my $page = shift;
  if (exists $seen{$page} && $sub2go ne "edit") {
    # this content has already been displayed
    return;
  }
  else {
    $seen{$page} = 1;
    open T, "<$config->{data}/$page" or die "Could not read page $page: $!\n";
    my $content = join "", <T>;
    close T;
    &fetch_toc($content);
    return $content;
  }
}

sub save {
  if ($input{preview}) {
    &edit;
    return;
  }
  my $content = $input{content};
  #### $content =~ s/\r//g;

  $rcs->file("$page");

  open T, ">$config->{data}/$page" or die "Could not write page $page: $!\n";
  print T $content;
  # FIXME ??? print T "\n";
  close T;

  if (! -e "$config->{data}/${page},v") {
    # initial checkin
    $rcs->ci('-u', '-mInitialCheckinByPodWiki');
    $rcs->co('-l');
  }
  else {
    my $user;
    $input{comment} ||= "*** empty log message ***";
    $input{login} = $current_user || "anonymous";
    $input{login} .= "-" . $ENV{REMOTE_ADDR};

    $rcs->ci('-l', "-m $input{comment}", "-w$input{login}");
  }

  # re-build the page index based on X<> indices
  $saving = 1;
  &tie_index;
  &render(&fetch_page($page));

  foreach my $attribute (keys %attributes) {
    # look for classification
    if (exists $input{"podattributes-${attribute}"}) {
      # ok, defined, classify the page
      &store_attribute($attribute, $input{"podattributes-${attribute}"});
    }
  }

  print qq(<p class="nolink">Saved successfully page <b>$page</b>.</p>);
}

sub interpolate {
  # interpolate embedded variables, if any
  my $content = shift;

  # interpolate internal variables
  $content =~ s/\$page/$page/g;
  $content =~ s/\$me/$me/g;
  $content =~ s/\$version/$PodWiki::VERSION/g;

  # interpolate user defined variables
  &tie_variables;
  foreach my $var (keys %variables) {
    $content =~ s/\$$var/$variables{$var}/g;
  }

  # interpolate masked \$ characters to $
  $content =~ s/\\\$/\$/g;

  return $content;
}

sub jump {
  # print a select box to jump directly to a page
  opendir D, "$config->{data}" or die "Could not read DATA directory $config->{data}: $!\n";
  print qq(<form name="jump" action="$me" method="get"><select name="page">);
  foreach my $page (sort grep { $_ !~ /^\./ } readdir(D)) {
    next if($page =~ /\,v$/);
    print qq(<option value="$page">$page</option>);
  }
  closedir D;
  print qq(</select><input type="submit" value="go"></form>);
}

sub info {
  $rcs->file("$page");
  my $revision = $rcs->head;
  my @revisions = $rcs->revisions;
  print qq(<h3>PageInfo about <img src="images/ilink.png">
  <i><a href="$me?page=$page">$page</a></i> (Revision: $revision)</h3>);

  my %comment = $rcs->comments;
  my $previous = 0;
  my %prev;
  foreach my $rev (reverse @revisions) {
    if ($previous) {
      $prev{$rev} = $previous;
    }
    $previous = $rev;
  }

  foreach my $rev (@revisions) {
    my $revdate = scalar localtime($rcs->revdate($rev));
    my $state   = $rcs->state($rev);
    my $author  = $rcs->author($rev);
    print qq(
     <table border="0" bgcolor="#000000" cellspacing="1" width=60%"><tr><td>
      <table border="0" cellspacing="1" cellpadding="5" width="100%">
      <tr>
        <th bgcolor="#cfcfcf" align="right" width="20%">Revision</th>
        <td bgcolor="#ffffff">
         $rev
         <a class="dotted" href="$me?page=$page&state=checkout&revision=$rev">Restore</a>
         <a class="dotted" href="$me?page=$page&state=view&revision=$rev">View</a>
    );

    if (exists $prev{$rev}) {
      print
	qq(<a class="dotted"
            href="$me?page=$page&state=diff&previous=$prev{$rev}&current=$rev">diff to $prev{$rev}</a>
      );
    }

    print qq(
        </td>
      </tr>
      <tr>
        <th bgcolor="#cfcfcf" align="right">Date</th><td bgcolor="#ffffff">$revdate</td>
      </tr>
      <tr>
        <th bgcolor="#cfcfcf" align="right">State</th><td bgcolor="#ffffff">$state</td>
      </tr>
      <tr>
        <th bgcolor="#cfcfcf" align="right">Author</th><td bgcolor="#ffffff">$author</td>
      </tr>
      <tr>
        <th bgcolor="#cfcfcf" align="right">Comment</th><td bgcolor="#ffffff">$comment{$rev}</td>
      </tr>
      </table></td></tr></table><p>&nbsp;</p>
    );
    $previous = $rev;
  }
}

sub pageindex {
  # build an index of all existing pages
  opendir D, "$config->{data}" or die "Could not read DATA directory $config->{data}: $!\n";
  print qq(
   <h3>PageIndex</h3>
   <table border="0" bgcolor="#000000" cellspacing="1"><tr><td>
    <table border="0" cellspacing="1" cellpadding="5" >
     <tr bgcolor="#cfcfcf">
      <th>Page</th><th>Revision</th><th>Author</th><th>Last modified</th><th>Size</th><th>Edit</th>
     </tr>
  );

  foreach my $page (sort grep { $_ !~ /^\./ } readdir(D)) {
    next if($page =~ /\,v$/);
    my $rcs = &rcsinit;
    $rcs->file($page);
    my $revision = $rcs->head;
    my $modified = scalar localtime($rcs->revdate);
    my $author   = $rcs->author;
    my $size     = (stat("$config->{data}/$page"))[7];
    my $link;
    if ($config->{direct}) {
      $link = "$page";
    }
    else {
      $link = "$me?page=$page";
    }
    print qq(
      <tr bgcolor="#ffffff">
       <td><p><img src="images/ilink.png"> <a href="$link">$page</a></p></td>
       <td><p>$revision</p></td>
       <td><p>$author</p></td>
       <td><p>$modified</p></td>
       <td><p>$size</p></td>
       <td>
         <p>
            <a class="dotted" href="$me?page=$page&state=edit">edit</a>&nbsp;
            <a class="dotted" href="$me?page=$page&state=remove">remove</a>
         </p>
       </td>
      </tr>
    );
  }
  print qq(</table></td></tr></table>);
  closedir D;

  # add image/file index here too
  &fileindex;
}


sub fileindex {
  ### images
  opendir D, "$config->{gfx}" or die "Could not read IMAGES directory $config->{gfx}: $!\n";
    print qq(
   <p>&nbsp;</p>
   <h3>FileIndex</h3>
   <table border="0" bgcolor="#000000" cellspacing="1"><tr><td>
    <table border="0" cellspacing="1" cellpadding="5" >
     <tr bgcolor="#cfcfcf">
      <th>Filename</th><th>Last modified</th><th>Edit</th>
     </tr>
  );

  my $tumb = "";
  foreach my $file (sort grep { $_ !~ /^\./ } readdir(D)) {
    my $modified = scalar localtime( (stat("$config->{gfx}/$file"))[9] );
    if ($config->{image_magick}) {
      # image magick support is enabled
      if ($file =~ /(jpg|jpeg|png|gif|bmp|xpm|tiff)$/) {  # FIXME: more suffixes to add?
	$tumb = qq(<img src="$me?state=imagetumb&image=$file" border="0"><br>);
      }
    }
    print qq(
      <tr bgcolor="#ffffff">
       <td align="center"><p>${tumb}${file}</p></td>
       <td><p>$modified</p></td>
       <td>
        <p>
         <a class="dotted" href="$config->{gfx}/$file" target="__new__">view</a>&nbsp;
         <a class="dotted" href="$me?file=$file&state=remove">remove</a>
        </p>
       </td>
      </tr>
    );
    $tumb = "";
  }
  print qq(</table></td></tr></table>);

  closedir D;
}

sub newest {
  opendir D, "$config->{data}" or die "Could not read DATA directory $config->{data}: $!\n";
  print "<h3>Newest Pages</h3>\n";
  print "<ul>\n";

  my %file;
  foreach my $page (sort grep { $_ !~ /^\./ } readdir(D)) {
    next if($page =~ /\,v$/);
    my $modified = (stat("$config->{data}/$page"))[9];
    $file{$page} = $modified;
  }

  my $counter = 5;
  foreach my $page (sort { $file{$b} <=> $file{$a} } keys %file) {
    if ($counter) {
      my $modified = scalar localtime($file{$page});
      print qq(
       <li><b>$modified</b>  <img src="images/ilink.png"> <a href="$me?page=$page">$page</a></li>
      );
    }
    else {
      last;
    }
    $counter--;
  }

  print "</ul>\n";
  closedir D;
}

sub list_userhomepages {
  &tie_users;
  print "<ul>\n";
  foreach my $user (sort keys %users) {
    my $home = "UserHome_${user}";
    my $link;
    if (-e "$config->{data}/$home") {
      $link = qq(<img src="images/ilink.png"> <a href="$me?page=$home">$user</a>);
    }
    else {
      $link = qq(<font class="nolink">$user</font><a href="$me?page=$home&state=edit">?</a>);
    }
    print "<li>$link</li>\n";
  }
  print "</ul>\n";
}

sub remove {
  print qq(<h3>Remove</h3>);
  if ($input{file}) {
    # image
    print qq(<p>Are you sure to remove the file <b>$input{file}</b>?</p>);
  }
  else {
    print qq(<p>Are you sure to remove the page <b>$page?</b></p>);
  }
  print qq(
   <form name="remove" action="$me" method="post">
    <input type="hidden" name="state" value="doremove">
    <input type="hidden" name="file" value="$input{file}">
    <input type="hidden" name="page" value="$page">
    <input type="submit" name="submit" value="YES">
    <input type="submit" name="submit" value="NO">
   </form>
  );
}

sub doremove {
  if ($input{submit} eq "NO") {
    # cancel
    &pageindex;
  }
  else {
    if ($input{file}) {
      # image
      unlink "$config->{gfx}/$input{file}" or die "Could not remove file $input{file}: $!\n";
      &pageindex;
      print qq(<p class="nolink">Removed successfully file <b>$input{file}</b>.</p>);
    }
    else {
      # page
      unlink "$config->{data}/$page" or die "Could not remove page $page: $!\n";
      unlink "$config->{data}/${page},v" or die "Could not remove page $page: $!\n";
      &pageindex;
      print qq(<p class="nolink">Removed successfully page <b>$page</b>.</p>);
    }
  }
}

sub search {
  print qq(
   <h3>Page Search</h3>
    <form name="search" action="$me" method="post">
     <input type="hidden" name="state" value="dosearch">
     <input type="text" name="string" size="30">
     <input type="submit" name="submit-full" value="full text search">
  );

  if (-e "$config->{data}/.podindex") {
    print qq(<input type="submit" name="submit-index" value="index search">);
  }

  my $a_size = scalar keys %attributes;
  if ($a_size) {
    # attributes exists, so provide them
    print "<p>&nbsp;</p><h3>Attribute Search</h3>";
    print qq(<table border="0"><tr valign="top">);
    my $sel;
    foreach my $attribute (sort keys %attributes) {
      next if $attribute =~ /^\./;
      print qq(<td><b>$attribute</b><br>);
      print qq(<select name="podattributes-${attribute}" size="10"><option></option>);
      foreach my $attr_value (sort split /\s\s*/, $attributes{$attribute}) {
	print qq(<option>$attr_value</option>);
      }
      print qq(</select></td>);
    }
    print qq(</tr></table>);
    print qq(<input type="submit" name="submit-attributes" value="attribute search">);
  }

  print qq(</form>);
}

sub dosearch {
  if (exists $input{"submit-full"}) {
    print qq(<h3>full text search results</h3>);
    opendir D, "$config->{data}" or die "Could not read DATA directory $config->{data}: $!\n";
    my @res;
    my $tree = new Pod::Tree;
    print "<dl>\n";
    foreach my $page (sort grep { $_ !~ /^\./ } readdir(D)) {
      next if($page =~ /,v$/);
      if (-T "$config->{data}/$page") {
	open F, "<$config->{data}/$page" or die "Could not read page $page: $!\n";
	my $content = join "", <F>;
	close F;
	$tree->load_string($content);
	my $plain = $tree->get_root->get_deep_text;
	if ($plain =~ /$input{string}/i) {
	  print "<dt><img src=\"images/ilink.png\"> <a href=\"$me?page=$page\">$page</a></dt>\n";
	  my ($front, $match, $back) = $plain =~ /(.{0,30})($input{string})(.{0,30})/i;
	  print "<dd>$front<b>$match<\/b>$back<p>&nbsp;</p><\/dd>\n";
	}
      }
    }
    print "</dl>\n";
    closedir D;
  }
  elsif (exists $input{"submit-index"}) {
    print qq(<h3>index search results</h3>);
    &tie_index;
    my %match;
    foreach my $word (split /\s\s*/, lc($input{string})) {
      if (exists $index{$word}) {
	map { $match{$_}++ } keys %{$index{$word}};
      }
    }
    foreach my $matchpage (sort { $match{$b} <=> $match{$a} } keys %match) {
      print "<p><img src=\"images/ilink.png\"> <a href=\"$me?page=$matchpage\">$matchpage</a></p>\n";
    }
  }
  elsif (exists $input{"submit-attributes"}) {
    print qq(<h3>attribute search results</h3>);
    my %match;
    foreach my $attribute (sort keys %attributes) {
      next if $attribute =~ /^\./;
      if ($input{"podattributes-${attribute}"}) {
	if (exists $attributes{".classified"}->{$attribute}->{ $input{"podattributes-${attribute}"} }) {
	  map { $match{$_}++ } keys %{$attributes{".classified"}->{$attribute}->{ $input{"podattributes-${attribute}"} }};
	}
      }
    }
    foreach my $matchpage (sort { $match{$b} <=> $match{$a} } keys %match) {
      print "<p><img src=\"images/ilink.png\"> <a href=\"$me?page=$matchpage\">$matchpage</a></p>\n";
    }
  }
}

sub upload {
  my $fh = $input{file};
  open F, "> $config->{gfx}/$input{file}" or die "Could not write file $config->{gfx}/$input{file}: $!\n";
  binmode F;
  my $buffer;
  while (read($fh,$buffer,1024)) {
    print F $buffer;
  }
  close F;
  &edit;
  print "<p>File $input{file} successfully uploaded.</p>\n";
}

sub checkout {
  # check out a revision of a page
  $rcs->file($page);
  unlink "$config->{data}/$page";
  $rcs->co("-r$input{revision}");
  $rcs->ci('-l', "-mRESTORED");
  &display;

  print "<p>Revision $input{revision} of page <b>$page</b> successfully restored.</p>\n";
}

sub diff {
  # print a diff between two revisions of a page
  print "<p><div class=pdotted>";
  $rcs->file($page);
  if (! $rcs->rcsdiff("-r$input{previous}", "-r$input{current}")) {
    print "<p>no changes between revision $input{previous} and $input{current}</p>";
  }
  else {
    my (undef, undef, @diff) = $rcs->rcsdiff("-r$input{previous}", "-r$input{current}", "-u");
    print qq(<table border="0" width="100%" cellspacing="0">);
    local $_;
    my $color;
    foreach (@diff) {
      next if /^\@\@/;
      if (/^\+/) {
	$color = "bgcolor=\"lightgreen\"";
      }
      elsif (/^\-/) {
	$color = "bgcolor=\"orange\"";
      }
      else {
	$color = "bgcolor=\"gray70\"";
      }
      s/\</&lt;/g;
      print qq(<tr><td $color><pre>&nbsp;$_</pre></td></tr>);
    }
    print qq(</table>);
  }
  print "</div>\n";
  print qq(<table border="0" width="10%">
    <tr><td bgcolor="lightgreen">added</td></tr>
    <tr><td bgcolor="orange">removed</td></tr>
    <tr><td bgcolor="gray70">unchanged</td></tr>
    </table>
  );

  &info;
}

sub view {
  # check out a temporary revision of a page and display it
  my $dir = "/tmp/.podwiki";
  system("mkdir -p $dir") and die "Could not create temporary directory $dir: $!\n";
  system("cp $config->{data}/${page},v $dir/");
  system("chmod 644 $dir/${page},v");
  my $rcs = new Rcs;
  $rcs->bindir("$config->{rcsbindir}");
  $rcs->workdir("$dir");
  $rcs->rcsdir("$dir");
  $rcs->file("$page");
  $rcs->co("-r$input{revision}");
  open F, "$dir/$page" or die "Could not read temporary revsion $input{revision} of $page: $!\n";
  my $content = join "", <F>;
  close F;

  unlink "$dir/$page";
  unlink "$dir/${page},v";

  #### $content =~ s/\r//g;
  print qq(
    <h3>View Revision $input{revision} of page
        <img src="images/ilink.png"> <a href="$me?page=$page">$page</a>:
    </h3><div class="pdotted">);
  $output = "";
  &render($content);
  print $output;
  print "</div><p>&nbsp;</p>";

  &info;
}

sub login {
  # resent the login screen and preserve CGI variables
  print qq(
   <h3>$next{message}</h3>
   <form name="login" action="$me" method="post">
   <input type="hidden" name="state" value="check_login">
   <input type="hidden" name="nextstate" value="$next{state}">
   <input type="hidden" name="page" value="$page">
    <table border="0">
      <tr>
        <th align="right">Login</th>
	<td><input type="text" size="23" name="user"></td>
      </tr>
      <tr>
        <th align="right">Password</th>
	<td><input type="password" size="23" name="passwd"></td>
      </tr>
      <tr>
        <td>&nbsp;</td>
	<td>
           <input style="width: 150;" type="submit" name="submit" value="Login to PodWiki">
        </td>
      </tr>
     <!--
      <tr>
	<td><p>&nbsp;</p></td>
	<td>
          <p><a href="$me?state=drop_message&user=root">Forgot your password? Request help!</a></p> 
        </td>
      </tr>
      -->
    </table>
   </form>
  );
}

sub logout {
  &set_cookie($current_user, "X", "-d1");
}

sub change_passwd {
  my ($user, $message) = @_;
  if (!$user) {
    $user = $current_user;
  }
  if (! $message) {
    $message = "Change password";
  }
  if (! $user) {
    # no user to change at all
    print qq(<p class="nolink">Please <a href="$me?state=login">login</a> first.</p>);
    return;
  }
  my(undef, $pw, $mail) = &fetch_user($user);
  print qq(
   <h3>$message</h3>
   <form name="change_passwd" action="$me" method="post">
   <input type="hidden" name="state" value="do_change_passwd">
    <table border="0">
      <tr>
        <th align="right">User</th>
	<td><input type="text" size="20" name="user" value="$user" readonly></td>
      </tr>
      <tr>
        <th align="right">Email</th>
	<td><input type="text" size="20" name="mail" value="$mail"></td>
      </tr>
      <tr>
        <th align="right">Old password</th>
	<td><input type="password" size="20" name="passwd_old"></td>
      </tr>
      <tr>
        <th align="right">New password</th>
	<td><input type="password" size="20" name="passwd_new"></td>
      </tr>
      <tr>
        <th align="right">New password - repeat</th>
	<td><input type="password" size="20" name="passwd_repeat"></td>
      </tr>
      <tr>
        <td>&nbsp;</td>
        <td><input type="submit" name="submit" value="Save Changes">
        <p class="pdotted">If you want to keep the current password, do not fill out the
        password fields!</p>
        </td>
      </tr>
    </table>
   </form>
  );
}

sub do_change_passwd {
  my($user, $md5pw, $mail) = &fetch_user($input{user});
  if (! $input{passwd_old} && $input{mail}) {
    # just change the password
    &save_user($user, "", $input{mail});
    if (-e "$config->{data}/UserHome_${$user}") {
      &render(&fetch_page("UserHome_${$user}"));
    }
    else {
      &render(&fetch_page("PodWikiIndex"));
    }
    print qq(<p class="nolink">Password changed successfully</p>);
  }
  else {
    # change password
    my $hashpw = md5_hex($input{passwd_old});
    my $newpw  = md5_hex($input{passwd_new});
    if (($md5pw eq $hashpw) && ($input{passwd_new} eq $input{passwd_repeat})) {
      &save_user($user, $input{passwd_new}, $input{mail});
      &render(&fetch_page("PodWikiIndex"));
      print qq(<p class="nolink">Password changed successfully</p>);

      # FIXME: maybe the user must be logged out here
      # FIXME: unclear: has the user to be logged in to be able to change password?
    }
    else {
      &change_passwd($user, "Change password - retry");
      print qq(<p class="nolink">Old password invalid or new passwords are not identical</p>);
    }
  }
}

### POD parser ###

sub render {
  my $input = &interpolate(shift());
  my $break_mode = "break";
  my $tree = new Pod::Tree;
  $tree->load_string($input);
  $tree->walk(\&walker);
}


sub walker {
  my $node = shift;
  my $type = $node->get_type;
  my $sub = "walk_" . $type;
  &$sub($node);
}


sub walk_root {
  my $node = shift;
  &walk_children($node);
}


sub walk_children {
  my $node     = shift;
  my $children = $node->get_children;

  foreach my $child (@$children) {
    &walker($child);
  }
}


sub walk_verbatim {
  my $node = shift;
  my $text = $node->get_text;
  print &tag_open($syntax{verbatim});
  $text =~ s/\s\s*$/ /gs;  # remove trailing spaces
  print $text;
  print &tag_close($syntax{verbatim});
  print "\n";
}


sub walk_ordinary {
  my $node = shift;
  print &tag_open($syntax{text});
  &walk_children($node);
  print &tag_close($syntax{text});
  print "\n";
}


sub walk_command {
  my $node    = shift;
  my $command = lc($node->get_command);
  if ($command =~ /(head[1-4])/) {
    my $target = &clean_head($node->get_deep_text);
    print "<a  name=\"$target\"></a>";
    print &tag_open($syntax{$command});
    &walk_children($node);
    print &tag_close($syntax{$command});
  }
  elsif ($command =~ /over/) {
    &walk_list($node);
  }
}

sub walk_letter {
  # workaround for node-type 'letter', which
  # is handled in walk_sequence()
  &walk_sequence(@_);
}

sub walk_sequence {
  my $node   = shift;
  my $letter = lc($node->get_letter);

  if ($letter =~ /i|b|c|f/) {
    # format element
    print &tag_open($syntax{$letter});
    &walk_children($node);
    print &tag_close($syntax{$letter});
  }
  elsif ($letter =~ /x/) {
    # index element
    if ($saving) {
      foreach my $word (split /\s\s*/, lc($node->get_deep_text)) {
	if (exists $index{$word}) {
	  # add current page to this word
	  my $H = $index{$word};
	  $H->{$page} = 1;
	  $index{$word} = $H;
	}
	else {
	  # add the word (along with current page)
	  $index{$word} = { $page => 1 };
	}
      }
    }
  }
  elsif ($letter =~ /p/) {
    # include another pod page here
    my $targetpage = $node->get_deep_text;
    if (-e "$config->{data}/$targetpage") {
      &render(&fetch_page($targetpage));
    }
    elsif ($targetpage =~ /^(intern|state):(.*)$/) {
      my ($target, $cgi_params) = split/\?/, $2;
      if ($target eq "state") {
	if ($sub2go) {
	  # internal state (== sub)
	  &$sub2go;
	}
	else {
	  # no state, means display a page
	  &render(&fetch_page($page));
	}
      }
      else {
	if ($cgi_params) {
	  my @pairs = split /&/, $cgi_params;
	  foreach my $pair (@pairs) {
	    my ($opt, $val) = split /=/, $pair;
	    $input{$opt} = $val;
	  }
	}
	&$target; # just run the internal sub, i.e. display or pageinfo
      }
    }
    elsif ($targetpage =~ /^http:\/\//i) {
      # fetch external content
      print &fetch_http($targetpage);
    }
    else {
      # new, non-existing page, present an "add this page" link
      print "P&lt;<font class=\"nolink\">$targetpage</font><a href=\"$me?page=$targetpage&state=edit&inherit=$page\">?</a>>";
    }
  }
  elsif ($letter =~ /g/) {
    # graphic element
    my ($image, $alt) = split/\|/, $node->get_deep_text;
    if ($image !~ /\//) {
      # podwiki-maintained image in data/ directory
      $image = "$config->{gfx}/$image";
    }
    print "<img src=\"$image\" border=\"0\" alt=\"$alt\">";
  }
  elsif ($letter =~ /s/) {
    # non-breakable html output
    $break_mode = "nobreak";
    &walk_children($node);
    $break_mode = "break";
  }
  elsif ($letter =~ /l/) {
    # link
    my($uri, $name);
    $uri = $name = "";
    my $target       = $node->get_target;
    my $linkpage     = $target->get_page;
    ($uri, $name) = split /\s\s*/, $linkpage, 2;
    my $section      = $target->get_section;

    if ($uri =~ /^(http|https|ftp|mailto):/) {
      # external uri
      print "<img src=\"images/elink.png\"> <a href=\"$uri";
    }
    elsif ($uri =~ /^(state|intern):(.*)$/) {
      # internal state uri
      my ($state, $cgi_params) = split/\?/, $2;
      print "<img src=\"images/ilink.png\"> <a href=\"$me?state=" . lc($state);
      if ($cgi_params) {
	print "&$cgi_params";
      }
    }
    else {
      # internal uri (podwiki word)
      if (-e "$config->{data}/$linkpage") {
	# linkpage exists
	print "<img src=\"images/ilink.png\"> <a href=\"";
	if ($config->{direct}) {
	  print "$uri";
	}
	else {
	  print "$me?page=$uri";
	}
      }
      else {
	# nope
	$name = "?";
	print "<font class=\"nolink\">$uri</font><a href=\"$me?page=$uri&state=edit&inherit=$page";
      }
    }
    if ($section) {
      # anchor reference
      print "#" . $section;
    }
    print "\">";
    if ($name) {
      print $name;
    }
    else {
      &walk_children($node);
    }
    print "</a>";
  }
  elsif ($letter =~ /e/) {
    # character entity, html unicode and pod entities are supported
    my $entity = lc($node->get_deep_text);
    print "&" .$entity . ";";
  }
}


sub walk_text {
  my $node  = shift;
  my $text  = $node->get_text;
  $text =~ s/\s\s*$/ /gs;  # remove trailing spaces
  if ($break_mode eq "nobreak") {
    $text =~ s/\s/\&nbsp;/g;
  }
  print $text;
}


sub walk_list {
  my $node      = shift;
  my $indent    = $node->get_arg;
  my $list_type = $node->get_list_type;

  print &tag_open($syntax{"list_" . $list_type});

  &walk_children($node);	# text of the =over paragraph

  print &tag_close($syntax{"list_" . $list_type});
}


sub walk_item {
  my $node      = shift;
  my $item_type = $node->get_item_type;

  if ($item_type eq "text") {
    # indended text, handle different than ordinary lists
    print "<dt>";
    &walk_children($node);	# text of the =item paragraph
    print "</dt>";
    print &tag_open($syntax{"item_" . $item_type});
    &walk_siblings($node);
    print &tag_close($syntax{"item_" . $item_type});
  }
  else {
    # ordinary list
    print &tag_open($syntax{"item_" . $item_type});
    if ($item_type ne "bullet") {
      # bullet types do not have content on the =item line beside the bullet
      &walk_children($node);	# text of the =item paragraph
      print "<br>";
    }
    &walk_siblings($node);
    print &tag_close($syntax{"item_" . $item_type});
  }
}


sub walk_for {
  my $node      = shift;
  my $formatter = $node->get_arg;
  my $text      = $node->get_text;
  if ($formatter =~ /html/i) {
    print $text;  # 1:1 HTML code
  }
  elsif ($formatter =~ /txt|text|pre/i) {
    print "<pre>$text</pre>"; # 1:1 txt content
  }
  elsif ($formatter =~ /perl/i) {
    if ($config->{enable_perl_formatter}) {
      eval "$text"; # evaluate perl code
    }
    else {
      print "&lt;enable_perl_formatter::OFF>";
    }
  }
}


sub walk_siblings {
  my $node     = shift;
  my $siblings = $node->get_siblings;

  for my $sibling (@$siblings) {
    &walker($sibling);
  }
}


sub tag_open {
  my $name = shift;
  return "<" . $name . ">";
}


sub tag_close {
  my $name = shift;
  return "</" . $name . ">";
}


sub data_empty {
  opendir D, "$config->{data}" or die "Could not read DATA directory $config->{data}: $!\n";
  my @entries = grep { $_ !~ /^\./ } readdir(D);
  closedir D;
  my $num = scalar @entries;
  if ($num >= 1) {
    return 0; # not empty
  }
  else {
    return 1; # empty
  }
}

sub rcsinit {
  # create a rcs object
  my $rcs = new Rcs;
  $rcs->bindir("$config->{rcsbindir}");
  $rcs->workdir("$config->{data}");
  $rcs->rcsdir("$config->{data}");
  # $rcs->quiet(0);
  return $rcs;
}

sub setup {
  # initially set up things

  # do not execute setup if already done so
  return if(! &data_empty);

  # create the content
  opendir D, "$config->{templates}" or die "Could not read template directory: $!\n";
  foreach my $file (grep { $_ !~ /^\./ } readdir(D)) {
    system("cp $config->{templates}/$file $config->{data}/$file") and
      die "Could not install template file $file: $!\n";
    $rcs->file("$file");
    $rcs->ci('-u', '-mInitialCheckinByPodWikiSetup');
    $rcs->co('-l');
  }

  &tie_users;
  &tie_groups;
  &save_user($config->{default_admin_user}, $config->{default_admin_passwd});
  &set_member($config->{default_admin_user}, $config->{default_admin_group});
}


sub access_denied {
  # present the login box within normal page content
  my $message    = shift;
  $next{state}   = $sub2go;
  $next{message} = $message;
  $sub2go        = "login";
  return;
}

sub check_auth {
  if (! &data_empty) {
    &parse_options($page);
    my(@wr_groups, @rd_groups);
    (@wr_groups) = split(/\s*,\s*/, $page_conf{write}) if(exists $page_conf{write});
    (@rd_groups) = split(/\s*,\s*/, $page_conf{read}) if(exists $page_conf{read});
    if ($sub2go =~ /^edit|save|remove|doremove|checkout$/) {
      if ($page_conf{write}) {
	# user wants to edit the page but it is
	# write protected check credentials
	if (! &check_credencials(@wr_groups) ) {
	  &access_denied("write access denied");
	}
      }
      else {
	($current_user, undef) = &fetch_session;
      }
    }
    elsif ($sub2go eq "check_login") {
      # step in, check supplied credentials and build a cookie
      # session for this user, then call check_auth() again
      # do not attempt to check membership because this info
      # is not available yet
      if (! &check_credencials() ) {
	$sub2go = $input{nextstate};
	&access_denied("login failed");
      }
      else {
	# access granted
	if (! $input{nextstate}) {
	  # direct login
	  if (-e "$config->{data}/UserHome_$input{user}") {
	    $page = "UserHome_$input{user}";
	  }
	  $sub2go = "";
	}
	else {
	  $sub2go = $input{nextstate};
	}
      }
    } # if(check_login)
    elsif ($sub2go eq "logout") {
      &set_cookie($current_user, "X", "-d1");
      $sub2go = "";
    }
    elsif ($sub2go =~ /_admin$/) {
      # root permissions required!
      if (! &check_credencials("root")) {
	&access_denied("root access denied");
      }
      # root access granted
    }
    else {
      if ($page_conf{read}) {
	# page is not publicly readable
	# check credentials
	if (! &check_credencials(@rd_groups) ) {
	  &access_denied("read access denied");
	}
      }
      else {
	# unprotected, just deliver the page
	($current_user, undef) = &fetch_session;
	return;
      }
    }
  } # if(data_empty)
}

sub tie_variables {
  if (! tied %variables) {
    tie %variables, "DB_File", "$config->{data}/.podvariables" or die "Could not open .podvariables file: $!\n";
  }
}

sub tie_users {
  if (! tied %users) {
    tie %users, "DB_File", "$config->{data}/.podusers" or die "Could not open .podusers file: $!\n";
  }
}

sub tie_groups {
  if (! tied %groups) {
    tie %groups, "DB_File", "$config->{data}/.podgroups" or die "Could not open .podgroups file: $!\n";
  }
}

sub tie_attributes {
  if (! tied %attributes) {
    tie %attributes, "MLDBM", "$config->{data}/.podattributes" or die "Could not open .podattributes file: $!\n";
  }
}

sub tie_index {
  if (! tied %index) {
    tie %index, "MLDBM", "$config->{data}/.podindex" or die "Could not open .podindex file: $!\n";
  }
}

sub store_attribute {
  my($attribute, $value) = @_;
  my $H = $attributes{".classified"};
  foreach my $attr_value (keys %{$H->{$attribute}}) {
    # remove from other values, if any
    delete $H->{$attribute}->{$attr_value}->{$page};
  }
  if ($value) {
    # only add it if something has been selected
    $H->{$attribute}->{$value}->{$page} = 1;
  }
  $attributes{".classified"} = $H;
}


sub fetch_user {
  my $user = shift;
  &tie_users;
  if (exists $users{$user}) {
    my($pw, $mail) = split /:/, $users{$user};
    return ($user, $pw, $mail);
  }
  else {
    return (0,0);
  }
}

sub save_user {
  my ($user, $passwd, $mail) = @_;
  &tie_users;
  if (! exists $users{$user}) {
    &create_user_home($user);
  }
  my(undef, $p, $m) = &fetch_user($user);
  if (! $passwd) {
    $passwd = $p;
  }
  else {
    $passwd = md5_hex($passwd);
  }
  if (! $mail) {
    $mail = $m;
  }
  $users{$user} = $passwd . ":" . $mail;
}

sub create_user_home {
  my $user = shift;
  my $home = "UserHome_${user}";
  $rcs->file("$home");

  open T, ">$config->{data}/$home" or die "Could not write page $home: $!\n";
  print T "=head3 UserHome of B<$user>\n\n=cut\n";
  close T;

  if (! -e "$config->{data}/${home},v") {
    # initial checkin
    $rcs->ci('-u', '-mInitialCheckinByPodWiki');
    $rcs->co('-l');
  }
  else {
    my $user;
    $input{comment} = "*** empty log message ***";
    $input{login} = $current_user || getpwuid($<);
    $rcs->ci('-l', "-m $input{comment}", "-w$input{login}");
  }
}

sub is_member {
  my ($user, @groups) = @_;
  &tie_groups;
  foreach my $group (@groups) {
    if (exists $groups{$group}) {
      if (grep {$user eq $_} split /\s*,\s*/, $groups{$group} ) {
	# user *IS* a listed member
	return 1;
      }
    }
  }
  # not found
  return 0;
}

sub set_member {
  my($user, $group) = @_;
  &tie_groups;
  my @users = split /\s*,\s*/, $groups{$group};
  if (! grep {$user eq $_} @users) {
    $groups{$group} = join ",", (@users, $user);
  }
}

sub set_members {
  my($group, @users) = @_;
  &tie_groups;
  $groups{$group} = join ",", (@users);
}

sub del_member {
  my($user, $group) = @_;
  &tie_groups;
  if ($group eq "*") {
    # remove $user from all groups
    foreach my $group (keys %groups) {
      if (grep {$user eq $_} split /\s*,\s*/, $groups{$group}) {
	my %gusers = map {$_ => 1} split /\s*,\s*/, $groups{$group};
	delete $gusers{$user};
	$groups{$group} = join ",", keys %gusers;
      }
    }
  }
  else {
    # remove $user from just one group
    my %gusers = map {$_ => 1} split /\s*,\s*/, $groups{$group};
    delete $gusers{$user};
    $groups{$group} = join ",", keys %gusers;
  }
}

sub del_user {
  my ($user) = @_;
  &tie_users;
  delete $users{$user};
}

sub del_group {
  my ($group) = @_;
  &tie_groups;
  delete $groups{$group};
}

sub fetch_session {
  my %cookie = fetch CGI::Cookie;
  if (exists $cookie{session}) {
    # expected cookie exists, user seems to be already logged in
    my ($user, $hashpw) = split /:/, $cookie{session}->value();
    return ($user, $hashpw);
  }
  else {
    # no cookie
    return (0,0);
  }
}

sub check_credencials {
  my @groups = @_;
  my(%session, %cookie, $user, $hashpw, $md5pw, $mail);

  if (exists $input{user} && exists $input{passwd}) {
    # CGI login, do not use eventually existing cookie
    $user = $input{user};
    $hashpw = md5_hex($input{passwd});
  }
  else {
    # expected cookie exists, user seems to be already logged in
    ($user, $hashpw) = &fetch_session;
    if (! $user) {
      # no cookie
      return 0;
    }
  }

  # validate credentials(either supplied per cookie or CGI)
  ($user, $md5pw, $mail)  = &fetch_user($user);
  if ($hashpw eq $md5pw) {
    # success, user is authenticated
    &set_cookie($user, $hashpw);
    $current_user = $user;
    if (@groups) {
      # group info available, check membership
      if (&is_member($user, @groups)) {
	# and is a member in one of the configured groups
	return 1;
      }
      else {
	# no member
	return 0;
      }
    }
    else {
      # no group info, membership is not interesting for the moment,
      # grant access now
      return 1;
    }
  }
  else {
    # invalid passwd
    return 0;
  }
}

sub set_cookie {
  my($user, $hashpw, $expire) = @_;
  if (! $expire) {
    $expire = $config->{expire};
  }
  my $cookie = new CGI::Cookie(
			       -name    => "session",
			       -value   => "$user:$hashpw",
			       -expires => "$expire"
			       );
  print "Set-Cookie: $cookie\n";
}

sub change_root_passwd {
  &change_passwd("root", "Please change the <b>root</b> password!");
}

sub parse_options {
  # parse config options per page
  my $page = shift;
  if (-e "$config->{data}/$page") {
    if (! %page_conf) {
      # do it only if not already done
      open P, "$config->{data}/$page" or die "Could not open page $page for reading: $!\n";
      %page_conf = ();
      local $_;
      my $o;
      while (<P>) {
	next if(/^\s*$/);
	chomp;
	if (/^=(begin|for) options/) {
	  $o = 1;
	}
	elsif ($o) {
	  if (/^=(head|end|cur|over|begin)/) {
	    last;
	  }
	  else {
	    s/^\s*//g;
	    s/\s*$//g;
	    my($opt, $val) = split /\s*=\s*/;
	    $page_conf{$opt} = $val;
	  }
	}
      }
      close P;
    }
  }
  %page_conf;
}

sub fetch_http {
  my $uri = shift;
  if ($uri =~ /^http:\/\/(.*)$/) {
    my ($host, $path) = split /\//, $1, 2;
    my (undef, $port) = split /:/, $host, 2;
    $path = "/" . $path;
    $port ||= 80;
    my $conn = new IO::Socket::INET(
                                  Proto    => "tcp",
                                  PeerAddr => $host,
                                  PeerPort => $port,
                                 ) or die "Unable to connect to $host:$port: $!\n";
    $conn->autoflush(1);

    print $conn qq(GET $path HTTP/1.0
User-Agent: Mozilla/5.0 (compatible; PodWiki::fetch_http $PodWiki::VERSION)
Host: $host
Accept: text/*;q=1.0, image/png;q=1.0, image/jpeg;q=1.0, image/gif;q=1.0, image/*;q=0.8, */*;q=0.5
Accept-Charset: iso-8859-1;q=1.0, *;q=0.9, utf-8;q=0.8


);
    my $content = join "", <$conn>;
    close $conn;

    return  "<!-- START - external content included by intern:fetch_http. uri: $uri-->\n"
          . &clean_http_content($content)
	  . "<!--   END - external content included by intern:fetch_http. uri: $uri-->\n";
  }
}


sub clean_http_content {
  # FIXME: replace <img src=..> uri's with remote hostname
  my $content = shift;

  # remove header fields
  $content =~ s/\r//g;
  $content =~ s/^.+?\n\n//s;

  # remove redundant HTML tags
  $content =~ s/(<html>|<body[^>]*>|<\/html>|<\/body>)//gis;

  # remove HTML header field, if any. we are using our own
  # and this could disrupt our page design/layout/style
  $content =~ s/<head>.*<\/head>//is;

  return $content;
}


sub imagetumb {
  print "Content-type: image/png\n\n";
  binmode STDOUT;
  my (undef, undef, $size, @undef) = split / /, `$config->{image_magick}/identify $config->{gfx}/$input{image}`;
  $size =~ s/\+.*$//;
  my($x, $y) = split /x/, $size;

  if ($x > $config->{tumbsize} && $y > $config->{tumbsize}) {
    system("$config->{image_magick}/convert -resize $config->{tumbsize}x$config->{tumbsize} $config->{gfx}/$input{image} png:-")
      and die "Could not open pipe to $config->{image_magick}/convert: $!";
  }
  else {
    system("$config->{image_magick}/convert $config->{gfx}/$input{image} png:-")
      and die "Could not open pipe to $config->{image_magick}/convert: $!";
  }
}

sub dumper {
  print "<pre>" . Dumper(shift()) . "</pre>\n";
}

END {
  eval {
    untie %users;
    untie %groups;
  }; # ignore errors, if any
}

1; #end













